import sys

try:
    from importlib.resources import files  # ... and any other things you want to get
except ImportError:
    from importlib_resources import files  # pip install importlib_resources

module_path = files(sys.modules[__name__])
data_dirpath = module_path / 'data'

ppath = module_path.joinpath
dpath = data_dirpath.joinpath



from IPython.display import Javascript
from copy import copy
from typing import Optional

CHANNEL_TYPES = ['audio', 'data']


def single_time_vis(channel: dict,
                    bt=0,
                    tt=0,
                    chart_type: Optional[str] = None,
                    enable_playback: bool = True,
                    height: int = 50,
                    params=None,
                    title: str = '',
                    subtitle='',
                    **kwargs
                    ) -> Javascript:
    """
    Render a visualization of time series or audio data.

    :param channel: A time channel dict or a list of numbers (see below)
    :param bt: The logical start time of the display, in microseconds
    :param tt: The logical end time of the display, in microseconds
    :param chart_type: The type of visualization to draw, one of
        ["bargraph", "heatmap", "spectrogram", "peaks", or "winners"]
    :param enable_playback: Whether to embed audio controls in the component (audio channel only)
    :param height: The height of the channel in pixels
    :param params: An empty dict or a dict with "chunk_size" as an integer in milliseconds
    :param title: A title to display above the chart
    :param subtitle: A subtitle to display under the title

    Channel keys:

    Audio channel:
    Should include either "url" or both "wf" and "sr"

    :param url: A URL to access WAV-format audio over HTTP
    :param wf: A list of 16-bit integers
    :param sr: The sample rate of the recording

    Data channel:

    Can be a list of numbers (for bargraphs or heatmaps) or strings (for winners).
    Otherwise, a channel dict with the following keys:

    :param data: A list of data points, with the keys "value" and either ("bt" and "tt") or "time"
    :param bargraphMax: The numeric value for the top of the chart, default 1
    :param bargraphMin: The numeric value for the top of the chart, default -2
    :param categories: For a winners channel, the list of categories to display
        (matching the "winners" values of the data points)
    """
    channel = _preprocess_channel(channel)
    props = dict(kwargs, bt=bt, tt=tt, chart_type=chart_type, enable_playback=enable_playback, height=height,
                 params=params, title=title, subtitle=subtitle)
    return _single_time_vis(channel, props)


def time_vis(channels, props=None) -> Javascript:
    """
    Render multiple channels.

    :param channels: The channels to render, a list of dicts.
    :return:
    """
    if not props:
        props = {}
    channels = [_preprocess_channel(channel) for channel in channels]
    js_source = f'renderMultiTimeVis(element.get(0), {channels}, {props})'.replace('True', 'true').replace('None',
                                                                                                           'null')
    jsobj = Javascript(js_source)
    jsobj._trace = {'channels': channels, 'props': props}
    return jsobj


def _preprocess_channel(channel):
    preprocessed = copy(channel)
    if isinstance(preprocessed, list):
        preprocessed = {
            'chart_type': 'bargraph',
            'data': preprocessed,
            'type': 'data',
        }
    data = preprocessed.get('data', [])
    if data:
        if isinstance(data[0], str):
            preprocessed['chart_type'] = 'winners'
            preprocessed['categories'] = list(set(data))
        else:
            preprocessed['bargraphMax'] = max(*data)
    if 'wf' in preprocessed or 'url' in preprocessed:
        preprocessed['type'] = 'audio'
        preprocessed['chart_type'] = 'peaks'
    return preprocessed


def _single_time_vis(channel, props) -> Javascript:
    ##########################################################################################
    # TODO: Horrible hack due to horrible JS interface. Change!
    #   chartType is a visualization concern, and should therefore be in props, not channel
    #   (also, channel should probably be called src to align with new word for concept)
    if 'chart_type' in props:
        channel['chartType'] = props['chart_type']
    ##########################################################################################

    js_source = f'renderTimeChannel(element.get(0), {channel}, {props})'.replace('True', 'true').replace('None', 'null')
    jsobj = Javascript(js_source)
    jsobj._trace = {'channel': channel, 'props': props}
    return jsobj



from IPython.display import Javascript

js_libs = ['https://otosense-dev-ui.s3.amazonaws.com/static/js/chord_chart.js']


def chord_chart(data, options=None):
    """
    A chord chart for visualizing a confusion matrix.

    :param data: A CentroidSmoothing model.
    :return:
    """
    options = options or {}
    return Javascript(f"""
    ((element) => {{
        console.log('Drawing chord chart.');
        require(['chord_chart'], (chordChart) => chordChart(element.get(0), {data}))
    }})(element);""", lib=js_libs)



import io
import os
from pathlib import PurePath
from typing import Iterable
from oui.multi_time_vis.base import single_time_vis

CHANNEL_TYPES = ['audio', 'data']

DFLT_SR = 44100
DFLT_CHART_TYPE = 'spectrogram'
MAX_WF_LEN_FOR_SPECTROGRAMS = 44100 * 3600 / 2  # would be half an hour at 44100 Hz
DFLT_ENABLE_PLAYBACK = True
DFLT_HEIGHT = 100
DFLT_PARAMS = None


# The convenience function
def jsobj_of_audio(src,
                   chart_type=None,
                   enable_playback=DFLT_ENABLE_PLAYBACK,
                   height=DFLT_HEIGHT,
                   params=DFLT_PARAMS,
                   title=None,
                   subtitle='',
                   **kwargs):
    """Make a (jupyter displayable) jsobj from an audio source, flexibly.

    :param src: some recognized source of audio.
        At this point: waveform (list/array of int samples), (wf, sr), filepath, bytes, file-like object
    :param chart_type: The chart type to render, either 'peaks' or 'spectrogram'
    :param enable_playback: Whether to enable playback on double click (default True)
    :param height: The height of the chart in pixels (default 50)
    :param params: Extra rendering parameters, currently unused
    :param title: The title to display, defaults to the filename
    :param subtitle: An optional subtitle to display under the title
    :param kwargs: extra kwargs to be passed on to Javascript object constructor
    :return:
    """
    if isinstance(src, bytes):
        src = io.BytesIO(src)
    if isinstance(src, (str, io.IOBase, PurePath)):
        return file_to_jsobj(src, chart_type, enable_playback, height, params, title, subtitle, **kwargs)
    elif isinstance(src, Iterable):
        return wfsr_to_jsobj(src, chart_type, enable_playback, height, params, title, subtitle, **kwargs)
    else:
        raise TypeError(f"Unrecognized src type: {type(src)}")


def wfsr_to_src_spec(wf, sr=44100):
    duration_s = len(wf) / sr

    return {
        'type': 'audio',
        'wf': _cast_wf(wf),
        'sr': sr,
        'bt': 0,
        'tt': int(duration_s * 1000000)
    }


# The base function
def wfsr_to_jsobj(
        src,
        chart_type=None,
        enable_playback=DFLT_ENABLE_PLAYBACK,
        height=DFLT_HEIGHT,
        params=DFLT_PARAMS,
        title=None,
        subtitle='',
        **kwargs):
    """Make a (jupyter displayable) jsobj from a (waveform, sample rate)  or just waveform source

    :param src
    :param wf: Waveform. An iterable of ints
    :param sr: Sample rate. An int.
    :param chart_type: The chart type to render, either 'peaks' (default) or 'spectrogram'
    :param enable_playback: Whether to enable playback on double click (default True)
    :param height: The height of the chart in pixels (default 50)
    :param params: Extra rendering parameters, currently unused
    :param title: The title to display, defaults to the filename
    :param subtitle: An optional subtitle to display under the title
    :param kwargs: extra kwargs to be passed on to Javascript object constructor
    :return:
    """
    if isinstance(src, tuple) and len(src) == 2:  # then assume it's a (wf, sr):
        wf, sr = src
    else:
        wf = src
        sr = DFLT_SR
    if chart_type is None:
        if len(wf) > MAX_WF_LEN_FOR_SPECTROGRAMS:
            chart_type = 'peaks'
        else:
            chart_type = 'spectrogram'
    src_spec = wfsr_to_src_spec(wf, sr)
    title = title or ''
    return single_time_vis(src_spec,
                           bt=src_spec['bt'],
                           tt=src_spec['tt'],
                           chart_type=chart_type,
                           enable_playback=enable_playback,
                           height=height,
                           params=params,
                           title=title,
                           subtitle=subtitle,
                           **kwargs)


# A function for file-like sources
def file_to_jsobj(src,
                  chart_type=DFLT_CHART_TYPE,
                  enable_playback=DFLT_ENABLE_PLAYBACK,
                  height=DFLT_HEIGHT,
                  params=DFLT_PARAMS,
                  title=None,
                  subtitle='',
                  **kwargs
                  ):
    """Renders a time visualization of a WAV file from its file.

    :param src: The filepath str (or posix path) or file-like object (e.g. open file, or BytesIO object)
    :param chart_type: The chart type to render, either 'peaks' (default) or 'spectrogram'
    :param enable_playback: Whether to enable playback on double click (default True)
    :param height: The height of the chart in pixels (default 50)
    :param params: Extra rendering parameters, currently unused
    :param title: The title to display, defaults to the filename
    :param subtitle: An optional subtitle to display under the title
    :param kwargs: extra kwargs to be passed on to Javascript object constructor
    """
    import soundfile

    wfsr = soundfile.read(src, dtype='int16')
    if title is None and isinstance(src, str):
        title = os.path.basename(src)
    return wfsr_to_jsobj(wfsr,
                         chart_type=chart_type,
                         enable_playback=enable_playback,
                         height=height,
                         params=params,
                         title=title,
                         subtitle=subtitle,
                         **kwargs
                         )


render_wav_file = file_to_jsobj  # back-compatibility alias


def _cast_wf(wf):
    """Cast wf to a list of ints"""
    if not isinstance(wf, list):
        if str(type(wf)) == "<class 'numpy.ndarray'>":
            # see https://stackoverflow.com/questions/2060628/reading-wav-files-in-python
            wf = wf.tolist()  # list(wf) does not convert int16 to int
        else:
            wf = list(wf)  # fallback
    if len(wf) > 0:
        assert isinstance(wf[0], int), f"first element of wf wasn't an int, but a {type(wf[0])}"
    return wf



class ModuleNotFoundIgnore:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ModuleNotFoundError:
            pass
        return True


class ModuleNotFoundManager:
    """(Context) managing ModuleNotFoundErrors (ignore, log, or other creative uses...)

    Specifying no arguments will result in ignoring whole block silently

    >>> with ModuleNotFoundManager():
    ...     import i_do_not_exist

    Specifying a message will result in calling the log_func (default is ``print``) with msg as the argument.

    >>> with ModuleNotFoundManager("That module does not exist"):
    ...     import i_do_not_exist.either
    That module does not exist

    """

    def __init__(self, msg=None, log_func=print):
        self.msg = msg
        self.log_func = log_func

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is ModuleNotFoundError:
            if self.msg is not None:
                self.log_func(self.msg)
        return True



from oui.multi_time_vis.test import dpath


def audio_jsobj_are_equivalent(jsobj_01, jsobj_02):
    return (
            jsobj_01._trace['props'] == jsobj_02._trace['props']
            and
            jsobj_02._trace['channel'] == jsobj_01._trace['channel']
    )


def test_jsobj_of_audio():
    from oui.multi_time_vis import jsobj_of_audio
    import soundfile as sf

    posix_path = dpath('baby_voice.wav')
    jsobj_from_posix_path = jsobj_of_audio(posix_path)

    # test with filepath (string) input
    filepath = str(posix_path)
    jsobj_from_filepath = jsobj_of_audio(filepath)
    jsobj_from_filepath._trace['props']['title'] = ''  # hack removing automatic title that happens for string inputs
    assert (audio_jsobj_are_equivalent(jsobj_from_filepath, jsobj_from_posix_path))

    # test with bytes input
    b = posix_path.read_bytes()
    jsobj_from_bytes = jsobj_of_audio(b)
    assert (audio_jsobj_are_equivalent(jsobj_from_bytes, jsobj_from_posix_path))

    # test with an io.BytesIO object
    from io import BytesIO
    bb = BytesIO(b)
    jsobj_from_bytesio = jsobj_of_audio(bb)
    assert (audio_jsobj_are_equivalent(jsobj_from_bytesio, jsobj_from_posix_path))

    # test with (wf, sr) pair input
    wf, sr = sf.read(filepath, dtype='int16')
    jsobj_from_wfsr = jsobj_of_audio((wf, sr))
    assert (audio_jsobj_are_equivalent(jsobj_from_wfsr, jsobj_from_posix_path))

    # if sr is the DFLT_SR, test with (wf, sr) pair input
    from oui.multi_time_vis.audio import DFLT_SR
    if sr == DFLT_SR:
        jsobj_from_wf = jsobj_of_audio(wf)
        assert (audio_jsobj_are_equivalent(jsobj_from_wf, jsobj_from_posix_path))



from oui.multi_time_vis.base import (
    single_time_vis,
    time_vis,
    CHANNEL_TYPES
)

from oui.multi_time_vis.audio import (
    jsobj_of_audio,
    wfsr_to_jsobj,
    file_to_jsobj,
    render_wav_file,  # deprecated alias
)


import os
import json
from pathlib import Path

from IPython.display import display, Javascript

pkg_root_dir = os.path.dirname(__file__)
pjoin = lambda *p: os.path.join(pkg_root_dir, *p)
js_filename = pjoin('js', 'index.js')
with open(js_filename) as js_file:
    js_source = js_file.read()
    display(Javascript(js_source))

data_dirpath = pjoin('data')
djoin = lambda *p: os.path.join(data_dirpath, *p)


def get_pkg_data(filename):
    path = djoin(filename)
    if filename.endswith('.json'):
        return json.load(open(path, 'r'))
    else:
        raise ValueError(f"Unrecognized extension in {path}")

